@@ -7,22 +7,21 @@ package diagnostics
 import (
 	"context"
 	"fmt"
+	"os"
 	"os/exec"
 	"testing"
 
-	"golang.org/x/tools/gopls/internal/hooks"
 	"golang.org/x/tools/gopls/internal/protocol"
 	"golang.org/x/tools/gopls/internal/server"
 	. "golang.org/x/tools/gopls/internal/test/integration"
 	"golang.org/x/tools/gopls/internal/test/integration/fake"
 	"golang.org/x/tools/gopls/internal/util/bug"
-	"golang.org/x/tools/gopls/internal/util/goversion"
 	"golang.org/x/tools/internal/testenv"
 )
 
 func TestMain(m *testing.M) {
 	bug.PanicOnBugs = true
-	Main(m, hooks.Options)
+	os.Exit(Main(m))
 }
 
 // Use mod.com for all go.mod files due to golang/go#35230.
@@ -1385,36 +1384,6 @@ func _() {
 	})
 }
 
-func TestEnableAllExperiments(t *testing.T) {
-	// Before the oldest supported Go version, gopls sends a warning to upgrade
-	// Go, which fails the expectation below.
-	testenv.NeedsGo1Point(t, goversion.OldestSupported())
-
-	const mod = `
--- go.mod --
-module mod.com
-
-go 1.12
--- main.go --
-package main
-
-import "bytes"
-
-func b(c bytes.Buffer) {
-	_ = 1
-}
-`
-	WithOptions(
-		Settings{"allExperiments": true},
-	).Run(t, mod, func(t *testing.T, env *Env) {
-		// Confirm that the setting doesn't cause any warnings.
-		env.OnceMet(
-			InitialWorkspaceLoad,
-			NoShownMessage(""), // empty substring to match any message
-		)
-	})
-}
-
 func TestSwig(t *testing.T) {
 	if _, err := exec.LookPath("swig"); err != nil {
 		t.Skip("skipping test: swig not available")
